#include "bmi088.h"

bmi088_err_t bmi088_read(bmi088_t* dev, uint8_t address, uint8_t reg, uint8_t* buf, size_t len)
{
  i2c_t* i2c = dev->bus;
  i2c->start();
  i2c->write(address << 1);
  i2c->write(reg);
  i2c->stop();
  i2c->start();
  i2c->write(address << 1 | 1);
  for(size_t i=0 ; i < len - 1; ++i)
  {
    buf[i] = i2c->read(i2c->ack);
  }
  buf[len - 1] = i2c->read(i2c->nak);
  i2c->stop();
}

bmi088_err_t bmi088_write(bmi088_t* dev, uint8_t address, uint8_t reg, uint8_t* buf, size_t len)
{
  i2c_t* i2c = dev->bus;
  i2c->start();
  i2c->write(address << 1);
  i2c->write(reg);
  for(size_t i=0 ; i < len; ++i)
  {
    i2c->write(buf[i]);
  }
  i2c->stop();
}

bool bmi088_selftest(bmi088_t* dev)
{
  // Conduct self-test according to datasheet section 4.6
  return true;
}

void bmi088_setup_samplerate(bmi088_t* dev, bmi088_acc_bwp acc_bwp, bmi088_acc_odr_t acc_odr, bmi088_gyr_odr_t gyr_odr)
{
  uint8_t value = acc_bwp << 4 | acc_odr;
  bmi088_write(dev, dev->acc_address, BMI088_REG_ACC_CONF, &value, 1);
  value = gyr_odr;
  bmi088_write(dev, dev->gyr_address, BMI088_REG_GYR_BANDWIDTH, &value, 1);
}

bool bmi088_acc_drdy(bmi088_t* dev)
{
  uint8_t value;
  bmi088_read(dev, dev->acc_address, BMI088_REG_ACC_STATUS, &value, 1);
  return (value & 0x80) > 0;
}

void bmi088_setup_range(bmi088_t* dev, bmi088_acc_range_t acc_range, bmi088_gyr_range_t gyr_range)
{
  uint8_t value = acc_range;
  bmi088_write(dev, dev->acc_address, BMI088_REG_ACC_RANGE, &value, 1);
  dev->acc_factor = pow(2.0, acc_range + 1) * 1.5 / 32768.0;
  value = gyr_range;
  bmi088_write(dev, dev->gyr_address, BMI088_REG_GYR_RANGE, &value, 1);
  switch(gyr_range)
  {
  case BMI088_GYR_RANGE_2000S:
    dev->gyr_factor = 1.0 / 32768.0 * 2000.0;
    break;
  case BMI088_GYR_RANGE_1000S:
    dev->gyr_factor = 1.0 / 32768.0 * 1000.0;
    break;
  case BMI088_GYR_RANGE_500S:
    dev->gyr_factor = 1.0 / 32768.0 * 500.0;
    break;
  case BMI088_GYR_RANGE_250S:
    dev->gyr_factor = 1.0 / 32768.0 * 250.0;
    break;
  case BMI088_GYR_RANGE_125S:
    dev->gyr_factor = 1.0 / 32768.0 * 125.0;
    break;
  }
}

bmi088_data_t bmi088_convert_raw(bmi088_t* dev, bmi088_data_raw_t* data)
{
  bmi088_data_t res;
  res.acc[0] = (float)data->acc[0] * dev->acc_factor;
  res.acc[1] = (float)data->acc[1] * dev->acc_factor;
  res.acc[2] = (float)data->acc[2] * dev->acc_factor;
  res.gyr[0] = (float)data->gyr[0] * dev->gyr_factor;
  res.gyr[1] = (float)data->gyr[1] * dev->gyr_factor;
  res.gyr[2] = (float)data->gyr[2] * dev->gyr_factor;
  return res;
}

bmi088_err_t bmi088_setup(bmi088_t* dev, i2c_t* bus, uint8_t acc_address, uint8_t gyr_address)
{
  bmi088_err_t err = BMI088_OK;
  uint8_t buf[4];
  uint8_t error;
  uint8_t chip_id;

  dev->bus = bus;
  dev->acc_address = acc_address;
  dev->gyr_address = gyr_address;
  bmi088_read(dev, dev->acc_address, BMI088_REG_ACC_ID, &chip_id, 1);
  if(chip_id != 0x1e)
  {
    err = BMI088_INIT_ERROR;
    goto exit;
  }
  bmi088_read(dev, dev->gyr_address, BMI088_REG_GYR_ID, &chip_id, 1);
  if(chip_id != 0x0f)
  {
    err = BMI088_INIT_ERROR;
    goto exit;
  }
  bmi088_read(dev, dev->acc_address, BMI088_REG_ACC_CONF, &chip_id, 1);
  //printf("imu:addr %02x, acc_conf: %02x\n", acc_address, chip_id);
  if(!bmi088_selftest(dev))
  {
    err = BMI088_SELFTEST_ERROR;
  }
  buf[0] = 0xB6;
  bmi088_write(dev, dev->acc_address, BMI088_REG_ACC_SOFTRESET, buf, 1);
  bmi088_write(dev, dev->gyr_address, BMI088_REG_GYR_SOFTRESET, buf, 1);
  _waitms(30);
  bmi088_setup_samplerate(dev, BMI088_ACC_BWP_NORMAL, BMI088_ACC_ODR_12_5, BMI088_GYR_ODR_1000_116);
  bmi088_setup_range(dev, BMI088_ACC_RANGE_24G, BMI088_GYR_RANGE_2000S);
  bmi088_read(dev, dev->acc_address, BMI088_REG_ACC_ERR, &error, 1);
  if(error)
  {
    err = BMI088_ERROR_REG;
    bmi088_read(dev, dev->acc_address, BMI088_REG_ACC_CONF, buf, 1);
    printf("imu error: %02x, conf: %02x\n", error, buf[0]);
    goto exit;
  }
  // Turn on acc
  buf[0] = 0x04;
  bmi088_write(dev, dev->acc_address, BMI088_REG_ACC_PWR_CTRL, buf, 1);
  _waitus(450);
exit:
  return err;
}
