#include <stdio.h>
#include <propeller2.h>

#include "canapi.h"

static long stack[128];
static can_controller_t controller;

static void print_frame(can_frame_t *f, uint32_t timestamp)
{
    uint8_t len = can_frame_get_data_len(f);
    printf(can_frame_is_extended(f) ? "0x%08x " : "0x%03x ", can_frame_get_arbitration_id(f));
    if (can_frame_is_remote(f)) {
        printf("R"); // Remote frame doesn't have any data, len will be zero
    }
    for (uint32_t i = 0; i < len; i++) {
        printf("%02x", can_frame_get_data(f)[i]);
    }
    printf(" (%d)\n", timestamp);
}

static void can_mainloop()
{
  can_errorcode_t rc;
  can_bitrate_t bitrate = {.profile = CAN_BITRATE_500K_75};
  mcp25xxfd_spi_bind_redqueen2(&controller.host_interface);
  while (true) {
    rc = can_setup_controller(&controller, &bitrate, CAN_NO_FILTERS, CAN_MODE_NORMAL, CAN_OPTIONS_NONE);
    if (rc != CAN_ERC_NO_ERROR) {
      // This can fail if the CAN transceiver isn't powered up properly. That might happen
      // if the board had 3.3V but not 5V (the transceiver needs 5V to operate).
      printf("CAN setup error: %d\n", rc);
      // Try again after 1 second
      _waitms(1000);
    }
    else {
      break;
    }
  }
  // Create a CAN frame with 11-bit ID of 0x123 and 5 byte payload of deadbeef00
  uint8_t data[5] = {0xdeU, 0xadU, 0xbeU, 0xefU, 0x00U};
  can_frame_t my_tx_frame;
  can_make_frame(&my_tx_frame, false, 0x123, sizeof(data), data, false);
  uint32_t queued_ok = 0;
  while (true) {
    if(mcp25xxfd_spi_gpio_irq_asserted(&controller.host_interface))
    {
      mcp25xxfd_irq_handler(&controller);
    }
    // Send our frame
    rc = can_send_frame(&controller, &my_tx_frame, false);
    if (rc != CAN_ERC_NO_ERROR) {
      // This can happen if there is no room in the transmit queue, which can
      // happen if the CAN controller is connected to a CAN bus but there are no
      // other CAN controllers connected and able to ACK a CAN frame, so the
      // transmit queue fills up and then cannot accept any more frames.
      printf("CAN send error: %d, sent=%d\n", rc, queued_ok);
    }
    else {
      queued_ok++;
      printf("sent: %d, %02x\n", queued_ok, can_frame_get_data(&my_tx_frame)[4]);
    }
    can_frame_get_data(&my_tx_frame)[4]++; // Update last byte of frame payload
    _waitms(250);
    // Print up to 10 received frames
    can_rx_event_t rx_event;
    uint32_t n = 0;
    while (n < 10U) {
      can_rx_event_t *e = &rx_event;
      if (can_recv(&controller, e) && can_event_is_frame(e)) {
        can_frame_t *f = can_event_get_frame(e);
        print_frame(f, e->timestamp);
        n++;
      }
      else {
        break; // No events left to check for this loop
      }
    }
  }
}

void can_start() {
  //can_mainloop();
  __builtin_cogstart(can_mainloop(), &stack[0]);
}
