#include "flight_sensors.h"

#include "p2-i2c.h"
#include <stdbool.h>
#include <sys/types.h>
#include <math.h>

#define I2C0_SCL 17
#define I2C0_SDA 18

#define I2C1_SCL 32
#define I2C1_SDA 33

#define IMU_A_ACC_ADDR 0x18
#define IMU_A_GYR_ADDR 0x68
#define IMU_B_ACC_ADDR 0x19
#define IMU_B_GYR_ADDR 0x69
#define BARO_A_ADDR 0x76
#define BARO_B_ADDR 0x77

#define MAG_A_ADDR 0x0E
#define MAG_B_ADDR 0x0F


i2c_t i2c0;
i2c_t i2c1;

#ifdef USE_IMU
bmi088_t imu_a;
bmi088_t imu_b;
#endif

#ifdef USE_BARO
ms5611_t baro_a;
ms5611_t baro_b;
#endif

#ifdef USE_MAG
#ifndef FAKE_MAG_A
bm1422_t mag_a;
#endif
bm1422_t mag_b;
#endif


static void bmi088_print_data(bmi088_data_t *imu_data)
{
  printf("acc: %f, %f, %f", imu_data->acc[0], imu_data->acc[1], imu_data->acc[2]);
  printf(", gyr: %f, %f, %f", imu_data->gyr[0], imu_data->gyr[1], imu_data->gyr[2]);
}

void flight_sensors_setup()
{
#ifdef USE_IMU
  bmi088_err_t bmi_err;
#endif
#ifdef USE_BARO
  ms5611_err_t baro_err;
#endif
#ifdef USE_MAG
  bm1422_err_t mag_err;
#endif

  i2c0.setup(I2C0_SCL, I2C0_SDA, 400, i2c0.PU_NONE);
  i2c1.setup(I2C1_SCL, I2C1_SDA, 400, i2c1.PU_NONE);
#ifdef USE_IMU
  bmi_err = bmi088_setup(&imu_a, &i2c0, IMU_A_ACC_ADDR, IMU_A_GYR_ADDR);
  if(bmi_err != BMI088_OK)
  {
    printf("error setting up imu_a\n");
  }
  bmi_err = bmi088_setup(&imu_b, &i2c0, IMU_B_ACC_ADDR, IMU_B_GYR_ADDR);
  if(bmi_err != BMI088_OK)
  {
    printf("error setting up imu_b\n");
  }
#endif
#ifdef USE_BARO
  baro_err = ms5611_setup(&baro_a, &i2c1, BARO_A_ADDR);
  if(baro_err != MS5611_OK)
  {
    printf("error setting up baro_a\n");
  }
  baro_err = ms5611_setup(&baro_b, &i2c1, BARO_B_ADDR);
  if(baro_err != MS5611_OK)
  {
    printf("error setting up baro_b\n");
  }
#endif
#ifdef USE_MAG
  #ifndef FAKE_MAG_A
  mag_err = bm1422_setup(&mag_a, &i2c1, MAG_A_ADDR);
  if(mag_err != BM1422_OK)
  {
    printf("error setting up mag_a, %d\n", mag_err);
  }
  else
  {
    bm1422_preamble(&mag_a, BM1422_SINGLE_SHOT, BM1422_14BIT);
  }
  #endif
  mag_err = bm1422_setup(&mag_b, &i2c1, MAG_B_ADDR);
  if(mag_err != BM1422_OK)
  {
    printf("error setting up mag_b, %d\n", mag_err);
  }
  else
  {
    bm1422_preamble(&mag_b, BM1422_SINGLE_SHOT, BM1422_14BIT);
  }
#endif
}

flight_sensors_data_t flight_sensors_read_data(flight_sensors_data_t* previous)
{
  flight_sensors_data_t res;
  res.timestamp = _getcnt();

#ifdef USE_BARO
  // Trigger conversion asap, as it takes time acconding to
  // page 3 in the datasheet. It's ~0.60ms / 256 OSR worst case
  ms5611_start_temp_conversion(&baro_a, MS5611_CMD_D2_OSR_256);
  ms5611_start_temp_conversion(&baro_b, MS5611_CMD_D2_OSR_256);
#endif
  res.segment_a = _getcnt() - res.timestamp;
#ifdef USE_MAG
  // Trigger also early, see page 6 of the datasheet 0.5ms
#ifndef FAKE_MAG_A
  bm1422_force(&mag_a);
#endif
  bm1422_force(&mag_b);
  res.segment_b = _getcnt() - res.timestamp;
#endif
#ifdef USE_IMU
  if(bmi088_acc_drdy(&imu_a))
  {
    bmi088_read(&imu_a, imu_a.acc_address, BMI088_REG_ACC_DATA, (uint8_t*)res.imu_a_raw_data.acc, 6);
  }
  if(bmi088_acc_drdy(&imu_b))
  {
    bmi088_read(&imu_b, imu_b.acc_address, BMI088_REG_ACC_DATA, (uint8_t*)res.imu_b_raw_data.acc, 6);
  }
  bmi088_read(&imu_a, imu_a.gyr_address, BMI088_REG_GYR_DATA, (uint8_t*)res.imu_a_raw_data.gyr, 6);
  bmi088_read(&imu_b, imu_b.gyr_address, BMI088_REG_GYR_DATA, (uint8_t*)res.imu_b_raw_data.gyr, 6);
#endif
  res.segment_c = _getcnt() - res.timestamp;
#ifdef USE_BARO
  res.temperature_a = ms5611_read_temp(&baro_a);
  res.temperature_a = ms5611_read_temp(&baro_b);
#endif
  res.segment_d = _getcnt() - res.timestamp;
#ifdef USE_BARO
  ms5611_start_pressure_conversion(&baro_a, MS5611_CMD_D1_OSR_256);
  ms5611_start_pressure_conversion(&baro_b, MS5611_CMD_D1_OSR_256);
#endif
  res.segment_e = _getcnt() - res.timestamp;
#ifdef USE_MAG
  // When all we do is read the mag, DRDY can't be asserted
  #if !defined(USE_IMU) && !defined(USE_BARO)
  _waitms(2);
  #endif
  if(previous)
  {
    if(bm1422_drdy(&mag_b))
    {
      res.mag_b = bm1422_single_shot_read(&mag_b);
    }
    else
    {
      printf("taking previous reading\n");
      res.mag_b = previous->mag_b;
    }
  }
  else
  {
    printf("wait_drdy\n");
    bm1422_wait_drdy(&mag_b);
    res.mag_b = bm1422_single_shot_read(&mag_b);
  }
#ifndef FAKE_MAG_A
  if(previous)
  {
    if(bm1422_drdy(&mag_a))
    {
      res.mag_a = bm1422_single_shot_read(&mag_a);
    }
    else
    {
      res.mag_a = previous->mag_a;
    }
  }
  else
  {
    bm1422_wait_drdy(&mag_a);
    res.mag_a = bm1422_single_shot_read(&mag_a);
  }
#else
  res.mag_a = res.mag_b;
#endif
  res.segment_f = _getcnt() - res.timestamp;
#endif
#ifdef USE_IMU
  res.imu_a_data = bmi088_convert_raw(&imu_a, &res.imu_a_raw_data);
  res.imu_b_data = bmi088_convert_raw(&imu_b, &res.imu_b_raw_data);
#else
  _waitms(2);
#endif
  res.segment_g = _getcnt() - res.timestamp;
#ifdef USE_BARO
  res.pressure_a = ms5611_read_pressure(&baro_a);
  res.pressure_b = ms5611_read_pressure(&baro_b);
#endif
  res.segment_h = _getcnt() - res.timestamp;
  res.duration = _getcnt() - res.timestamp;
  return res;
}
